@@ -19,8 +19,8 @@ import socket
 import getpass
 import re
 import threading
+import base64
 from binascii import hexlify
-from lxml import etree
 
 try:
     import selectors
@@ -28,17 +28,20 @@ except ImportError:
     import selectors2 as selectors
 
 from ncclient.capabilities import Capabilities
+from ncclient.logging_ import SessionLoggerAdapter
 
 import paramiko
 
-from ncclient.transport.errors import AuthenticationError, SessionCloseError, SSHError, SSHUnknownHostError
+from ncclient.transport.errors import AuthenticationError, SessionCloseError, SSHError, SSHUnknownHostError, NetconfFramingError
 from ncclient.transport.session import Session
 from ncclient.transport.session import NetconfBase
-from ncclient.xml_ import *
 
 import logging
 logger = logging.getLogger("ncclient.transport.ssh")
 
+PORT_NETCONF_DEFAULT = 830
+PORT_SSH_DEFAULT = 22
+
 BUF_SIZE = 4096
 # v1.0: RFC 4742
 MSG_DELIM = "]]>]]>"
@@ -72,13 +75,15 @@ def default_unknown_host_cb(host, fingerprint):
     """
     return False
 
+
 def _colonify(fp):
     fp = fp.decode('UTF-8')
     finga = fp[:2]
-    for idx  in range(2, len(fp), 2):
+    for idx in range(2, len(fp), 2):
         finga += ":" + fp[idx:idx+2]
     return finga
 
+
 if sys.version < '3':
     def textify(buf):
         return buf
@@ -92,7 +97,6 @@ else:
     from io import BytesIO as StringIO
 
 
-
 class SSHSession(Session):
 
     "Implements a :rfc:`4742` NETCONF session over SSH."
@@ -121,8 +125,10 @@ class SSHSession(Session):
         self._message_list = []
         self._closing = threading.Event()
 
+        self.logger = SessionLoggerAdapter(logger, {'session': self})
+
     def _dispatch_message(self, raw):
-        logger.info("Received from %s session %s:\n%s", self.host, self.id, raw)
+        self.logger.info("Received:\n%s", raw)
         return super(SSHSession, self)._dispatch_message(raw)
 
     def _parse(self):
@@ -136,7 +142,7 @@ class SSHSession(Session):
         state across method calls and if a chunk has been read it will not be
         considered again."""
 
-        logger.debug("parsing netconf v1.0")
+        self.logger.debug("parsing netconf v1.0")
         buf = self._buffer
         buf.seek(self._parsing_pos10)
         if MSG_DELIM in buf.read().decode('UTF-8'):
@@ -154,7 +160,7 @@ class SSHSession(Session):
             if len(remaining) > 0:
                 # There could be another entire message in the
                 # buffer, so we should try to parse again.
-                logger.debug('Trying another round of parsing since there is still data')
+                self.logger.debug('Trying another round of parsing since there is still data')
                 self._parse10()
         else:
             # handle case that MSG_DELIM is split over two chunks
@@ -163,7 +169,7 @@ class SSHSession(Session):
                 self._parsing_pos10 = 0
 
     def _parse11(self):
-        
+
         """Messages are split into chunks. Chunks and messages are delimited
         by the regex #RE_NC11_DELIM defined earlier in this file. Each
         time we get called here either a chunk delimiter or an
@@ -171,8 +177,8 @@ class SSHSession(Session):
         data. If there is not enough data, we will wait for more. If a
         delimiter is found in the wrong place, a #NetconfFramingError
         will be raised."""
-        
-        logger.debug("_parse11: starting")
+
+        self.logger.debug("_parse11: starting")
 
         # suck in whole string that we have (this is what we will work on in
         # this function) and initialize a couple of useful values
@@ -180,23 +186,23 @@ class SSHSession(Session):
         data = self._buffer.getvalue()
         data_len = len(data)
         start = 0
-        logger.debug('_parse11: working with buffer of %d bytes', data_len)
+        self.logger.debug('_parse11: working with buffer of %d bytes', data_len)
         while True and start < data_len:
             # match to see if we found at least some kind of delimiter
-            logger.debug('_parse11: matching from %d bytes from start of buffer', start)
+            self.logger.debug('_parse11: matching from %d bytes from start of buffer', start)
             re_result = RE_NC11_DELIM.match(data[start:].decode('utf-8'))
             if not re_result:
 
                 # not found any kind of delimiter just break; this should only
                 # ever happen if we just have the first few characters of a
                 # message such that we don't yet have a full delimiter
-                logger.debug('_parse11: no delimiter found, buffer="%s"', data[start:].decode())
+                self.logger.debug('_parse11: no delimiter found, buffer="%s"', data[start:].decode())
                 break
 
             # save useful variables for reuse
             re_start = re_result.start()
             re_end = re_result.end()
-            logger.debug('_parse11: regular expression start=%d, end=%d', re_start, re_end)
+            self.logger.debug('_parse11: regular expression start=%d, end=%d', re_start, re_end)
 
             # If the regex doesn't start at the beginning of the buffer,
             # we're in trouble, so throw an error
@@ -210,7 +216,7 @@ class SSHSession(Session):
                 start += re_end
                 message = ''.join(self._message_list)
                 self._message_list = []
-                logger.debug('_parse11: found end of message delimiter')
+                self.logger.debug('_parse11: found end of message delimiter')
                 self._dispatch_message(message)
                 break
 
@@ -219,35 +225,34 @@ class SSHSession(Session):
                 # string that will tell us how many bytes past the end of
                 # where it was found that we need to have available to
                 # save the next chunk off
-                logger.debug('_parse11: found chunk delimiter')
+                self.logger.debug('_parse11: found chunk delimiter')
                 digits = int(re_result.group(1))
-                logger.debug('_parse11: chunk size %d bytes', digits)
+                self.logger.debug('_parse11: chunk size %d bytes', digits)
                 if (data_len-start) >= (re_end + digits):
                     # we have enough data for the chunk
                     fragment = textify(data[start+re_end:start+re_end+digits])
                     self._message_list.append(fragment)
                     start += re_end + digits
-                    logger.debug('_parse11: appending %d bytes', digits)
-                    logger.debug('_parse11: fragment = "%s"', fragment)
+                    self.logger.debug('_parse11: appending %d bytes', digits)
+                    self.logger.debug('_parse11: fragment = "%s"', fragment)
                 else:
                     # we don't have enough bytes, just break out for now
                     # after updating start pointer to start of new chunk
                     start += re_start
-                    logger.debug('_parse11: not enough data for chunk yet')
-                    logger.debug('_parse11: setting start to %d', start)
+                    self.logger.debug('_parse11: not enough data for chunk yet')
+                    self.logger.debug('_parse11: setting start to %d', start)
                     break
-                
+
         # Now out of the loop, need to see if we need to save back any content
         if start > 0:
-            logger.debug(
+            self.logger.debug(
                 '_parse11: saving back rest of message after %d bytes, original size %d',
                 start, data_len)
             self._buffer = StringIO(data[start:])
             if start < data_len:
-                logger.debug('_parse11: still have data, may have another full message!')
+                self.logger.debug('_parse11: still have data, may have another full message!')
                 self._parse11()
-        logger.debug('_parse11: ending')
-
+        self.logger.debug('_parse11: ending')
 
     def load_known_hosts(self, filename=None):
 
@@ -280,14 +285,27 @@ class SSHSession(Session):
         while self.is_alive() and (self is not threading.current_thread()):
             self.join(10)
 
+        if self._channel:
+            self._channel.close()
         self._channel = None
         self._connected = False
 
-
     # REMEMBER to update transport.rst if sig. changes, since it is hardcoded there
-    def connect(self, host, port=830, timeout=None, unknown_host_cb=default_unknown_host_cb,
-                username=None, password=None, key_filename=None, allow_agent=True,
-                hostkey_verify=True, look_for_keys=True, ssh_config=None, sock_fd=None):
+    def connect(
+            self,
+            host,
+            port                = PORT_NETCONF_DEFAULT,
+            timeout             = None,
+            unknown_host_cb     = default_unknown_host_cb,
+            username            = None,
+            password            = None,
+            key_filename        = None,
+            allow_agent         = True,
+            hostkey_verify      = True,
+            hostkey_b64         = None,
+            look_for_keys       = True,
+            ssh_config          = None,
+            sock_fd             = None):
 
         """Connect via SSH and initialize the NETCONF session. First attempts the publickey authentication method and then password authentication.
 
@@ -295,7 +313,7 @@ class SSHSession(Session):
 
         *host* is the hostname or IP address to connect to
 
-        *port* is by default 830, but some devices use the default SSH port of 22 so this may need to be specified
+        *port* is by default 830 (PORT_NETCONF_DEFAULT), but some devices use the default SSH port of 22 (PORT_SSH_DEFAULT) so this may need to be specified
 
         *timeout* is an optional timeout for socket connect
 
@@ -311,6 +329,8 @@ class SSHSession(Session):
 
         *hostkey_verify* enables hostkey verification from ~/.ssh/known_hosts
 
+        *hostkey_b64* only connect when server presents a public hostkey matching this (obtain from server /etc/ssh/ssh_host_*pub or ssh-keyscan)
+
         *look_for_keys* enables looking in the usual locations for ssh keys (e.g. :file:`~/.ssh/id_*`)
 
         *ssh_config* enables parsing of an OpenSSH configuration file, if set to its path, e.g. :file:`~/.ssh/config` or to True (in this case, use :file:`~/.ssh/config`).
@@ -370,55 +390,96 @@ class SSHSession(Session):
                 sock = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
             sock.settimeout(timeout)
 
-        t = self._transport = paramiko.Transport(sock)
-        t.set_log_channel(logger.name)
+        self._transport = paramiko.Transport(sock)
+        self._transport.set_log_channel(logger.name)
         if config.get("compression") == 'yes':
-            t.use_compression()
+            self._transport.use_compression()
 
+        if hostkey_b64:
+            # If we need to connect with a specific hostkey, negotiate for only its type
+            hostkey_obj = None
+            for key_cls in [paramiko.DSSKey, paramiko.Ed25519Key, paramiko.RSAKey, paramiko.ECDSAKey]:
+                try:
+                    hostkey_obj = key_cls(data=base64.b64decode(hostkey_b64))
+                except paramiko.SSHException:
+                    # Not a key of this type - try the next
+                    pass
+            if not hostkey_obj:
+                # We've tried all known host key types and haven't found a suitable one to use - bail
+                raise SSHError("Couldn't find suitable paramiko key class for host key %s" % hostkey_b64)
+            self._transport._preferred_keys = [hostkey_obj.get_name()]
+        elif self._host_keys:
+            # Else set preferred host keys to those we possess for the host
+            # (avoids situation where known_hosts contains a valid key for the host, but that key type is not selected during negotiation)
+            if port == PORT_SSH_DEFAULT:
+                known_hosts_lookup = host
+            else:
+                known_hosts_lookup = '[%s]:%s' % (host, port)
+            known_host_keys_for_this_host = self._host_keys.lookup(known_hosts_lookup)
+            if known_host_keys_for_this_host:
+                self._transport._preferred_keys = [x.key.get_name() for x in known_host_keys_for_this_host._entries]
+
+        # Connect
         try:
-            t.start_client()
-        except paramiko.SSHException:
-            raise SSHError('Negotiation failed')
+            self._transport.start_client()
+        except paramiko.SSHException as e:
+            raise SSHError('Negotiation failed: %s' % e)
 
-        # host key verification
-        server_key = t.get_remote_server_key()
-
-        fingerprint = _colonify(hexlify(server_key.get_fingerprint()))
+        server_key_obj = self._transport.get_remote_server_key()
+        fingerprint = _colonify(hexlify(server_key_obj.get_fingerprint()))
 
         if hostkey_verify:
-            known_host = self._host_keys.check(host, server_key)
-            if not known_host and not unknown_host_cb(host, fingerprint):
-                raise SSHUnknownHostError(host, fingerprint)
+            is_known_host = False
 
+            # For looking up entries for nonstandard (22) ssh ports in known_hosts
+            # we enclose host in brackets and append port number
+            if port == PORT_SSH_DEFAULT:
+                known_hosts_lookup = host
+            else:
+                known_hosts_lookup = '[%s]:%s' % (host, port)
+
+            if hostkey_b64:
+                # If hostkey specified, remote host /must/ use that hostkey
+                if(hostkey_obj.get_name() == server_key_obj.get_name() and hostkey_obj.asbytes() == server_key_obj.asbytes()):
+                    is_known_host = True
+            else:
+                # Check known_hosts
+                is_known_host = self._host_keys.check(known_hosts_lookup, server_key_obj)
+
+            if not is_known_host and not unknown_host_cb(host, fingerprint):
+                raise SSHUnknownHostError(known_hosts_lookup, fingerprint)
+
+        # Authenticating with our private key/identity
         if key_filename is None:
             key_filenames = []
         elif isinstance(key_filename, (str, bytes)):
-            key_filenames = [ key_filename ]
+            key_filenames = [key_filename]
         else:
             key_filenames = key_filename
 
         self._auth(username, password, key_filenames, allow_agent, look_for_keys)
 
-        self._connected = True # there was no error authenticating
+        self._connected = True      # there was no error authenticating
         self._closing.clear()
+
         # TODO: leopoul: Review, test, and if needed rewrite this part
         subsystem_names = self._device_handler.get_ssh_subsystem_names()
         for subname in subsystem_names:
-            c = self._channel = self._transport.open_session()
-            self._channel_id = c.get_id()
+            self._channel = self._transport.open_session()
+            self._channel_id = self._channel.get_id()
             channel_name = "%s-subsystem-%s" % (subname, str(self._channel_id))
-            c.set_name(channel_name)
+            self._channel.set_name(channel_name)
             try:
-                c.invoke_subsystem(subname)
+                self._channel.invoke_subsystem(subname)
             except paramiko.SSHException as e:
-                logger.info("%s (subsystem request rejected)", e)
+                self.logger.info("%s (subsystem request rejected)", e)
                 handle_exception = self._device_handler.handle_connection_exceptions(self)
                 # Ignore the exception, since we continue to try the different
                 # subsystem names until we find one that can connect.
-                #have to handle exception for each vendor here
+                # have to handle exception for each vendor here
                 if not handle_exception:
                     continue
-            self._channel_name = c.get_name()
+            self._channel_name = self._channel.get_name()
             self._post_connect()
             return
         raise SSHError("Could not open connection, possibly due to unacceptable"
@@ -432,24 +493,25 @@ class SSHSession(Session):
             for cls in (paramiko.RSAKey, paramiko.DSSKey, paramiko.ECDSAKey):
                 try:
                     key = cls.from_private_key_file(key_filename, password)
-                    logger.debug("Trying key %s from %s" %
-                              (hexlify(key.get_fingerprint()), key_filename))
+                    self.logger.debug("Trying key %s from %s",
+                                      hexlify(key.get_fingerprint()),
+                                      key_filename)
                     self._transport.auth_publickey(username, key)
                     return
                 except Exception as e:
                     saved_exception = e
-                    logger.debug(e)
+                    self.logger.debug(e)
 
         if allow_agent:
             for key in paramiko.Agent().get_keys():
                 try:
-                    logger.debug("Trying SSH agent key %s" %
-                                 hexlify(key.get_fingerprint()))
+                    self.logger.debug("Trying SSH agent key %s",
+                                      hexlify(key.get_fingerprint()))
                     self._transport.auth_publickey(username, key)
                     return
                 except Exception as e:
                     saved_exception = e
-                    logger.debug(e)
+                    self.logger.debug(e)
 
         keyfiles = []
         if look_for_keys:
@@ -476,13 +538,13 @@ class SSHSession(Session):
         for cls, filename in keyfiles:
             try:
                 key = cls.from_private_key_file(filename, password)
-                logger.debug("Trying discovered key %s in %s" %
-                          (hexlify(key.get_fingerprint()), filename))
+                self.logger.debug("Trying discovered key %s in %s",
+                                  hexlify(key.get_fingerprint()), filename)
                 self._transport.auth_publickey(username, key)
                 return
             except Exception as e:
                 saved_exception = e
-                logger.debug(e)
+                self.logger.debug(e)
 
         if password is not None:
             try:
@@ -490,7 +552,7 @@ class SSHSession(Session):
                 return
             except Exception as e:
                 saved_exception = e
-                logger.debug(e)
+                self.logger.debug(e)
 
         if saved_exception is not None:
             # need pep-3134 to do this right
@@ -502,14 +564,14 @@ class SSHSession(Session):
         chan = self._channel
         q = self._q
 
-        def start_delim(data_len): return '\n#%s\n'%(data_len)
+        def start_delim(data_len): return '\n#%s\n' % (data_len)
 
         try:
             s = selectors.DefaultSelector()
             s.register(chan, selectors.EVENT_READ)
-            logger.debug('selector type = %s', s.__class__.__name__)
+            self.logger.debug('selector type = %s', s.__class__.__name__)
             while True:
-                    
+
                 # Will wakeup evey TICK seconds to check if something
                 # to send, more quickly if something to read (due to
                 # select returning chan in readable list).
@@ -530,27 +592,29 @@ class SSHSession(Session):
                         # End of session, unexpected
                         raise SessionCloseError(self._buffer.getvalue())
                 if not q.empty() and chan.send_ready():
-                    logger.debug("Sending message")
+                    self.logger.debug("Sending message")
                     data = q.get()
                     if self._base == NetconfBase.BASE_11:
                         data = "%s%s%s" % (start_delim(len(data)), data, END_DELIM)
                     else:
                         data = "%s%s" % (data, MSG_DELIM)
-                    logger.debug("Sending: %s", data)
+                    self.logger.info("Sending:\n%s", data)
                     while data:
                         n = chan.send(data)
                         if n <= 0:
                             raise SessionCloseError(self._buffer.getvalue(), data)
                         data = data[n:]
         except Exception as e:
-            logger.debug("Broke out of main loop, error=%r", e)
+            self.logger.debug("Broke out of main loop, error=%r", e)
             self._dispatch_error(e)
             self.close()
 
     @property
     def host(self):
         """Host this session is connected to, or None if not connected."""
-        return self._host
+        if hasattr(self, '_host'):
+            return self._host
+        return None
 
     @property
     def transport(self):
